package cn.com.qmth.examcloud.core.oe.task.service.job;

import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureQueueRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamCaptureQueueEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureQueueService;
import cn.com.qmth.examcloud.core.oe.task.service.bean.FaceApiParam;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import com.fasterxml.jackson.databind.JsonNode;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * 处理人脸抓拍照片比对任务
 */
@Component
public class FaceVerifyJobHandler {

    private static final Logger log = LoggerFactory.getLogger(FaceVerifyJobHandler.class);

    @Autowired
    private ExamCaptureQueueRepo examCaptureQueueRepo;

    @Autowired
    private ExamCaptureQueueService examCaptureQueueService;

    public void run(int shardTotal, int shardIndex, String jobParam) throws Exception {
        long startTime = System.currentTimeMillis();
        FaceApiParam param = this.parseJobParam(jobParam);

        // 根据任务调度策略，推荐50
        final int batchSize = 50;

        // 优先取“已交卷”的考试记录ID集合
        List<Long> examRecordDataIds = examCaptureQueueService.findQueuesGroupByExamRecordDataId(shardTotal,
                shardIndex, batchSize, param.getMaxErrorNum(), true);
        int firstCount = examRecordDataIds.size(), moreCount = 0;

        if (firstCount < batchSize) {
            // 未取到或取到的数量不足时，再按“不区分交卷状态”方式取一次考试记录ID集合来补充（避免调度任务不饱和空转）
            List<Long> moreExamRecordDataIds = examCaptureQueueService.findQueuesGroupByExamRecordDataId(shardTotal,
                    shardIndex, batchSize, param.getMaxErrorNum(), false);
            moreCount = moreExamRecordDataIds.size();
            examRecordDataIds.addAll(moreExamRecordDataIds);

            if (CollectionUtils.isEmpty(examRecordDataIds)) {
                log.warn("分片任务_人脸比对_{}_{} 暂无待处理记录！", shardTotal, shardIndex);
                return;
            }
        }

        Set<Long> todoIds = new HashSet<>(examRecordDataIds);
        List<ExamCaptureQueueEntity> todoQueues = examCaptureQueueRepo.findByExamRecordDataIdIn(todoIds);
        if (CollectionUtils.isEmpty(todoQueues)) {
            log.warn("分片任务_人脸比对_{}_{} 暂无待处理队列记录！", shardTotal, shardIndex);
            return;
        }

        long queryCost = System.currentTimeMillis() - startTime;
        log.warn("分片任务_人脸比对_{}_{} 本次处理考试记录数:{} 待比对照片数:{} firstCount:{} moreCount:{} cost:{}ms",
                shardTotal, shardIndex, todoIds.size(), todoQueues.size(), firstCount, moreCount, queryCost);
        Map<Long, List<ExamCaptureQueueEntity>> maps = todoQueues.stream().collect(Collectors.groupingBy(ExamCaptureQueueEntity::getExamRecordDataId));

        // 创建线程池
        ExecutorService executorService = Executors.newFixedThreadPool(param.getMaxThreadNum());
        AtomicInteger successCount = new AtomicInteger();
        for (Map.Entry<Long, List<ExamCaptureQueueEntity>> entry : maps.entrySet()) {
            executorService.execute(() -> {
                // 分别按“考试记录ID”集中处理该“考试记录ID”下的抓拍照片队列记录（通常只有几条）
                Long examRecordDataId = entry.getKey();
                final String lockKey = CacheConstants.LOCK_FACE_COMPARE + examRecordDataId;

                try {
                    SequenceLockHelper.getLockSimple(lockKey);
                    examCaptureQueueService.handlerExamCaptureQueuesByExamRecordDataId(examRecordDataId, entry.getValue(), param, successCount);
                } catch (Exception e) {
                    if (e instanceof SequenceLockException) {
                        // 若锁问题，下次会继续执行
                        log.warn("当前人脸比对任务获取锁失败！examRecordDataId:{}, redisKey:{}", examRecordDataId, lockKey);
                    } else {
                        // 若异常，下次会继续执行（需要排查原因）
                        log.error("当前人脸比对任务处理失败！examRecordDataId:{}, error:{}", examRecordDataId, e.getMessage());
                    }
                } finally {
                    SequenceLockHelper.releaseLockSimple(lockKey);
                }
            });
        }

        // 停止接受新任务，并等待所有任务完成
        executorService.shutdown();
        while (!executorService.isTerminated()) {
            // ignore
        }

        long allCost = Math.max((System.currentTimeMillis() - startTime) / 1000, 1);
        log.warn("分片任务_人脸比对_{}_{} 本次处理比对照片数:{} 成功数:{} 约{}个每秒 threadNum:{} cost:{}s",
                shardTotal, shardIndex, todoQueues.size(), successCount.get(), (float) todoQueues.size() / allCost,
                param.getMaxThreadNum(), allCost);

        // 清理大对象内存
        maps.clear();
        todoQueues.clear();
    }

    private FaceApiParam parseJobParam(String jobParam) {
        JsonNode jsonParams = new JsonMapper().getNode(jobParam);
        if (jsonParams == null) {
            return new FaceApiParam();
        }

        FaceApiParam param = new FaceApiParam();

        JsonNode maxThreadNum = jsonParams.get("maxThreadNum");
        if (maxThreadNum != null) {
            // 限制：最小值为1，最大值20
            int x = Math.max(maxThreadNum.asInt(), 1);
            param.setMaxThreadNum(Math.min(x, 20));
        }

        JsonNode maxErrorNum = jsonParams.get("maxErrorNum");
        if (maxErrorNum != null) {
            // 限制：最小值为1
            param.setMaxErrorNum(Math.max(maxErrorNum.asInt(), 1));
        }

        JsonNode useBaiduApi = jsonParams.get("useBaiduApi");
        if (useBaiduApi != null) {
            param.setUseBaiduApi(useBaiduApi.asBoolean(true));
        }

        JsonNode useLocalBaiduApiForFaceCompare = jsonParams.get("useLocalBaiduApiForFaceCompare");
        if (useLocalBaiduApiForFaceCompare != null) {
            param.setUseLocalBaiduApiForFaceCompare(useLocalBaiduApiForFaceCompare.asBoolean(true));
        }

        JsonNode useLocalBaiduApiForFaceLiveness = jsonParams.get("useLocalBaiduApiForFaceLiveness");
        if (useLocalBaiduApiForFaceLiveness != null) {
            param.setUseLocalBaiduApiForFaceLiveness(useLocalBaiduApiForFaceLiveness.asBoolean(true));
        }

        return param;
    }

}
